from PySide6.QtCore import QDate
from PySide6.QtCore import Qt
from PySide6.QtGui import QKeyEvent
from PySide6.QtWidgets import QHBoxLayout
from PySide6.QtWidgets import QLabel
from PySide6.QtWidgets import QLineEdit
from PySide6.QtWidgets import QPushButton
from PySide6.QtWidgets import QSizePolicy
from PySide6.QtWidgets import QWidget

from fwtv.widgets import settings_widget


class LoginWidget(QWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.qh = QHBoxLayout(self)

        self.fetch_data_label = QLabel("Fetch data", self)
        self.fetch_data_label.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        self.qh.addWidget(self.fetch_data_label)

        # defaults to fetching data from the last 2 months in order to not fetch entire data history
        self.start_picker = settings_widget.DateSettingWidget("From", QDate.currentDate().addMonths(-2), self)
        self.qh.addWidget(self.start_picker)

        self.end_picker = settings_widget.DateSettingWidget("Until", QDate.currentDate(), self)
        self.qh.addWidget(self.end_picker)

        self.label = QLabel("Enter api key", self)
        self.label.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        self.qh.addWidget(self.label)

        self.key = QLineEdit(self)
        self.key.setEchoMode(QLineEdit.EchoMode.Password)
        self.qh.addWidget(self.key)

        self.button = QPushButton("Fetch data", self)
        self.button.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        self.qh.addWidget(self.button)

        self.timeout = settings_widget.IntegerPickerWidget("Timeout", 5)
        self.qh.addWidget(self.timeout)

        self.setLayout(self.qh)

    def keyPressEvent(self, event: QKeyEvent):
        super().keyPressEvent(event)
        if event.key() == Qt.Key.Key_Return:
            self.button.clicked.emit()



import collections
import datetime

from factorialhr import models
from PySide6.QtWidgets import QVBoxLayout
from PySide6.QtWidgets import QWidget

from fwtv import verifier
from fwtv.widgets import settings_widget
from fwtv.widgets import table_widget


def get_errors(
    start: datetime.date,
    end: datetime.date,
    attendances: list[models.Attendance],
    employees: list[models.Employee],
    tolerance: int,
) -> tuple[dict[str, list[str]], dict[str, list[verifier.Error]]]:
    preconditions: dict[str, list[str]] = collections.defaultdict(list)
    employee_errors: dict[str, list[verifier.Error]] = collections.defaultdict(list)
    for employee in employees:
        name = employee.full_name
        employee_attendances: list[verifier.Attendance] = []

        for attendance in filter(lambda x: x.employee_id == employee.id, attendances):
            if not attendance.clock_in:
                preconditions[name].append(f'no clock in time provided for attendance with id "{attendance.id}"')
                continue
            clock_in = attendance.clock_in
            if clock_in.date() < start or clock_in.date() > end:
                continue
            if not attendance.clock_out:
                preconditions[name].append(f'no clock out time provided for clock in time "{clock_in}"')
                continue
            if not attendance.workable:
                continue  # it has been declared as a break
            try:
                # automated time tracking is not precise enough and also is not able to handle seconds precise enough
                a = verifier.Attendance(
                    clock_in=clock_in.replace(second=0), clock_out=attendance.clock_out.replace(second=0)
                )
                employee_attendances.append(a)
            except ValueError as e:
                preconditions[name].append(str(e))
                continue

        errors = verifier.verify_attendances(employee_attendances)
        # ignore all errors where the time attended is the tolerance above the limit, as factorial's automated time
        # tracking is not precises enough
        errors = [
            e
            for e in errors
            if e.time_attended > datetime.timedelta(hours=6, minutes=tolerance)
            and e.time_attended > datetime.timedelta(hours=9, minutes=tolerance)
        ]
        if errors:
            employee_errors[name] = errors
    return preconditions, employee_errors


class WorkingTimeWidget(QWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.teams = []
        self.attendances = []
        self.employees_by_team = {}
        self.employees = []
        self.qv = QVBoxLayout(self)

        self.settings_widget = settings_widget.SettingsWidget(self)
        self.qv.addWidget(self.settings_widget)

        self.preconditions_table = table_widget.PreconditionErrorsTableWidget(self)
        self.qv.addWidget(self.preconditions_table)

        self.failures_table = table_widget.FailuresTableWidget(self)
        self.qv.addWidget(self.failures_table)

        self.setLayout(self.qv)

        self.settings_widget.team_selector.selector.currentIndexChanged.connect(self.update_data)
        self.settings_widget.start_picker.date.dateChanged.connect(self.update_data)
        self.settings_widget.end_picker.date.dateChanged.connect(self.update_data)
        self.settings_widget.tolerance_selector.picker.valueChanged.connect(self.update_data)

        self.update_data()

    def set_data(
        self, teams: list[models.Team], attendances: list[models.Attendance], employees: list[models.Employee]
    ):
        self.attendances = attendances
        self.teams = teams
        self.employees = employees
        self.employees_by_team = {t.id: [e for e in employees if e.id in t.employee_ids] for t in teams}
        for i in range(self.settings_widget.team_selector.selector.count()):
            self.settings_widget.team_selector.selector.removeItem(i)
        self.settings_widget.team_selector.selector.addItems([team.name for team in teams])
        self.settings_widget.team_selector.selector.addItems([employee.full_name for employee in employees])
        self.update_data()

    def get_current_selection(self) -> models.Team | models.Employee | None:
        index = self.settings_widget.team_selector.selector.currentIndex()
        if 0 <= index < len(self.teams):
            return self.teams[index]
        elif 0 <= (index := index - len(self.teams)) < len(self.employees):
            return self.employees[index]
        else:
            return None

    def update_data(self):
        selection = self.get_current_selection()
        employees = []
        if isinstance(selection, models.Team):
            employees = self.employees_by_team[selection.id]
        elif isinstance(selection, models.Employee):
            employees = [selection]

        preconditions, errors = get_errors(
            self.settings_widget.start_picker.date.date().toPython(),
            self.settings_widget.end_picker.date.date().toPython(),
            self.attendances,
            employees,
            self.settings_widget.tolerance_selector.value(),
        )
        entries = collections.defaultdict(list)
        for k in preconditions.keys():
            entries[k] = [preconditions[k]]
        self.preconditions_table.set_data(entries)
        self.failures_table.set_data(errors)



from PySide6.QtCore import QDate
from PySide6.QtWidgets import QComboBox
from PySide6.QtWidgets import QDateEdit
from PySide6.QtWidgets import QDoubleSpinBox
from PySide6.QtWidgets import QHBoxLayout
from PySide6.QtWidgets import QLabel
from PySide6.QtWidgets import QSizePolicy
from PySide6.QtWidgets import QWidget


class TeamOrEmployeeSettingWidget(QWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.qh = QHBoxLayout(self)
        self.label = QLabel("Select a team or an employee", self)
        self.qh.addWidget(self.label)

        self.selector = QComboBox(self)
        self.selector.setEditable(True)
        self.selector.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        self.qh.addWidget(self.selector)

        self.setLayout(self.qh)


class DateSettingWidget(QWidget):
    def __init__(self, label: str, date: QDate, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.qh = QHBoxLayout(self)
        self.label = QLabel(label, self)
        self.qh.addWidget(self.label)

        self.date = QDateEdit(date, self)
        self.date.setCalendarPopup(True)
        self.date.setDisplayFormat("yyyy-MM-dd")
        self.qh.addWidget(self.date)

        self.setLayout(self.qh)


class IntegerPickerWidget(QWidget):
    def __init__(self, label: str, default: int, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.qh = QHBoxLayout(self)
        self.label = QLabel(label, self)
        self.qh.addWidget(self.label)

        self.picker = QDoubleSpinBox(self)
        self.picker.setSingleStep(1)
        self.picker.setDecimals(0)
        self.picker.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)
        self.picker.setValue(default)
        self.qh.addWidget(self.picker)

        self.setLayout(self.qh)

    def value(self) -> int:
        return int(self.picker.value())


class SettingsWidget(QWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.qh = QHBoxLayout(self)
        self.team_selector = TeamOrEmployeeSettingWidget(self)
        self.qh.addWidget(self.team_selector)

        last_month = QDate.currentDate().addMonths(-1)
        self.start_picker = DateSettingWidget("Start on", QDate(last_month.year(), last_month.month(), 1), self)
        self.qh.addWidget(self.start_picker)

        self.end_picker = DateSettingWidget(
            "End on", QDate(last_month.year(), last_month.month(), last_month.daysInMonth()).addDays(1), self
        )
        self.qh.addWidget(self.end_picker)

        self.tolerance_selector = IntegerPickerWidget("Tolerance", 1)
        self.qh.addWidget(self.tolerance_selector)

        self.setLayout(self.qh)



import sys

from factorialhr import endpoints
from PySide6.QtWidgets import QApplication
from PySide6.QtWidgets import QMessageBox
from PySide6.QtWidgets import QVBoxLayout
from PySide6.QtWidgets import QWidget

import fwtv
from fwtv.objects import async_converter
from fwtv.widgets import login_widget
from fwtv.widgets import working_time_widget


class MainWindow(QWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setWindowTitle(f"Factorial working time verification - version {fwtv.__version__}")
        self.qv = QVBoxLayout()
        self.login = login_widget.LoginWidget(self)
        self.qv.addWidget(self.login)

        self.verification_widget = working_time_widget.WorkingTimeWidget(self)
        self.qv.addWidget(self.verification_widget)

        self.setLayout(self.qv)

        self.login.button.clicked.connect(async_converter.ToAsync(self.fetch_data))

    async def fetch_data(self):
        self.login.button.hide()
        async with endpoints.NetworkHandler(self.login.key.text()) as api:
            try:
                _attendances = await endpoints.AttendanceEndpoint(api).all(
                    date_from=self.login.start_picker.date.date().toPython(),
                    date_to=self.login.end_picker.date.date().toPython(),
                    timeout=self.login.timeout.value(),
                )
                _employees = await endpoints.EmployeesEndpoint(api).all()
                _teams = await endpoints.TeamsEndpoint(api).all()
            except Exception as e:
                message_box = QMessageBox(self)
                message_box.setIcon(QMessageBox.Icon.Critical)
                message_box.setText(f"{type(e).__name__}\n{e}")
                message_box.setStandardButtons(QMessageBox.StandardButton.Ok)
                message_box.setDefaultButton(QMessageBox.StandardButton.Ok)
                message_box.exec()
                return
            finally:
                self.login.button.show()
        self.verification_widget.set_data(_teams, _attendances, _employees)


def main() -> int:
    app = QApplication(sys.argv)
    window = MainWindow()
    window.resize(1200, 675)
    window.show()
    return app.exec()


if __name__ == "__main__":
    sys.exit(main())



import traceback
import typing

import outcome
import trio
from PySide6.QtCore import QEvent
from PySide6.QtCore import QObject
from PySide6.QtWidgets import QApplication


class ToAsync(QObject):
    class ReenterQtObject(QObject):
        def event(self, event):
            if event.type() == QEvent.User + 1:
                event.fn()
                return True
            return False

    class ReenterQtEvent(QEvent):
        def __init__(self, fn):
            super().__init__(QEvent.Type(QEvent.User + 1))
            self.fn = fn

    def __init__(self, signal: typing.Callable):
        super().__init__()
        self.reenter_qt = self.ReenterQtObject()
        self.entry = signal

    def __call__(self, *args, **kwargs):
        trio.lowlevel.start_guest_run(
            self.entry,
            *args,
            run_sync_soon_threadsafe=self.next_guest_run_schedule,
            done_callback=self.trio_done_callback,
        )

    def next_guest_run_schedule(self, fn):
        QApplication.postEvent(self.reenter_qt, self.ReenterQtEvent(fn))

    @staticmethod
    def trio_done_callback(outcome_):
        if isinstance(outcome_, outcome.Error):
            error = outcome_.error
            traceback.print_exception(type(error), error, error.__traceback__)



import collections

import tabulate
from PySide6.QtCore import Qt
from PySide6.QtGui import QClipboard
from PySide6.QtGui import QKeyEvent
from PySide6.QtWidgets import QHeaderView
from PySide6.QtWidgets import QLabel
from PySide6.QtWidgets import QSizePolicy
from PySide6.QtWidgets import QTableView
from PySide6.QtWidgets import QTableWidget
from PySide6.QtWidgets import QTableWidgetItem
from PySide6.QtWidgets import QVBoxLayout
from PySide6.QtWidgets import QWidget

from fwtv import verifier


class ErrorTable(QTableWidget):
    def __init__(self, labels: list[str], *args, **kwargs):
        self.labels = labels
        super().__init__(0, len(self.labels), *args, **kwargs)
        self.setHorizontalHeaderLabels(self.labels)
        self.setSelectionBehavior(QTableView.SelectionBehavior.SelectRows)
        self._set_data({})

    def _set_data(self, data: dict[str, list[list[str]]]):
        self.setSortingEnabled(False)
        self.setRowCount(sum(len(failures) for failures in data.values()))
        row_index = 0
        for name, errors in data.items():
            for error in errors:
                self._set_row(row_index, name, *error)
                row_index += 1
        self.horizontalHeader().resizeSections(QHeaderView.ResizeMode.ResizeToContents)
        self.setSortingEnabled(True)

    def _set_row(self, row_index: int, *columns: str):
        for column_index, text in enumerate(columns):
            column_item = QTableWidgetItem(text)
            column_item.setFlags(
                Qt.ItemFlag.ItemNeverHasChildren
                | Qt.ItemFlag.ItemIsAutoTristate
                | Qt.ItemFlag.ItemIsEnabled
                | Qt.ItemFlag.ItemIsSelectable
            )
            self.setItem(row_index, column_index, column_item)

    def keyPressEvent(self, event: QKeyEvent):
        super().keyPressEvent(event)
        if event.key() == Qt.Key.Key_C and (event.modifiers() & Qt.KeyboardModifier.ControlModifier):
            row_by_index = collections.defaultdict(list)
            for index in self.selectedIndexes():
                row_by_index[index.row()].append(index.data())
            rows = []
            for row in row_by_index.values():
                rows.append(row)
            table = tabulate.tabulate(rows, headers=self.labels, tablefmt="orgtbl")
            QClipboard().setText(table)


class PreconditionErrorsTableWidget(ErrorTable):
    def __init__(self, *args, **kwargs):
        super().__init__(["Name", "Error"], *args, **kwargs)
        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

    def set_data(self, data: dict[str, list[list[str]]]):
        super()._set_data(data)
        self.horizontalHeader().setSectionResizeMode(1, QHeaderView.ResizeMode.Stretch)
        if not any(data) and not self.isHidden():
            self.hide()
        elif any(data) and self.isHidden():
            self.show()


class AdditionalInformationWidget(QWidget):
    def __init__(self, name: str, error: verifier.Error, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.qv = QVBoxLayout(self)

        self.label = QLabel(name)
        self.qv.addWidget(self.label)

        self.attendances = QTableWidget(len(error.attendances), 2, self)
        for i, a in enumerate(error.attendances):
            column_item = QTableWidgetItem(a.clock_in.strftime("%d.%m %H:%M"))
            column_item.setFlags(
                Qt.ItemFlag.ItemNeverHasChildren | Qt.ItemFlag.ItemIsAutoTristate | Qt.ItemFlag.ItemIsEnabled
            )
            self.attendances.setItem(i, 0, column_item)
            column_item = QTableWidgetItem(a.clock_out.strftime("%d.%m %H:%M"))
            column_item.setFlags(
                Qt.ItemFlag.ItemNeverHasChildren | Qt.ItemFlag.ItemIsAutoTristate | Qt.ItemFlag.ItemIsEnabled
            )
            self.attendances.setItem(i, 1, column_item)
        self.attendances.setHorizontalHeaderLabels(["Clock in", "Clock out"])
        self.attendances.horizontalHeader().resizeSections(QHeaderView.ResizeMode.ResizeToContents)
        self.qv.addWidget(self.attendances)

        self.error = QLabel(error.reason)
        self.qv.addWidget(self.error)

        self.setLayout(self.qv)


class FailuresTableWidget(ErrorTable):
    def __init__(self, *args, **kwargs):
        super().__init__(
            ["Name", "Affected Day(s)", "Reason", "Cumulated Break", "Cumulated Attendance"], *args, **kwargs
        )
        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        self.data: dict[str, list[verifier.Error]] = dict()
        self.cellDoubleClicked.connect(self.display_additional_information)

    def set_data(self, data: dict[str, list[verifier.Error]]):
        self.data = data
        entries = collections.defaultdict(list)
        for name, error in data.items():
            for e in error:
                affected_days = [str(day) for day in e.days_affected]
                affected_days.sort()
                entries[name].append([", ".join(affected_days), e.reason, str(e.break_time), str(e.time_attended)])
        super()._set_data(entries)
        self.horizontalHeader().setSectionResizeMode(2, QHeaderView.ResizeMode.Stretch)

    def display_additional_information(self, row: int, _: int):
        rows: list[tuple[str, verifier.Error]] = []
        for name, error in self.data.items():
            for e in error:
                rows.insert(0, (name, e))  # instead of reversing
        self.new_window = AdditionalInformationWidget(rows[row][0], rows[row][1])
        self.new_window.show()
